# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np

import pyttb as ttb

from .pyttb_utils import *


class symktensor(object):
    """
    SYMKTENSOR Class for symmetric Kruskal tensors (decomposed).

    """

    def __init__(self, *args):  # pragma:no cover
        assert False, "SYMKTENSOR class not yet implemented"
